# import the required libraries
import numpy as np
import matplotlib as mpl
import matplotlib.pyplot as plt

def lego_plot(xAmplitudes,yAmplitudes,nBins,xLabel,yLabel,xymind,xymaxd,title):

    x = np.array(xAmplitudes)   #turn x,y data into numpy arrays
    y = np.array(yAmplitudes)   #useful for regular matplotlib arrays

    fig = plt.figure()          #create a canvas, tell matplotlib it's 3d
    ax = fig.add_subplot(111, projection='3d')

    #make histogram stuff - set bins - I choose 20x20 because I have a lot of data
    #hist, xedges, yedges = np.histogram2d(x, y, bins=(20,20))
    hist, xedges, yedges = np.histogram2d(x, y, bins=(nBins,nBins))
    xpos, ypos = np.meshgrid(xedges[:-1]+xedges[1:], yedges[:-1]+yedges[1:])

    xpos = xpos.flatten()/2.
    ypos = ypos.flatten()/2.
    zpos = np.zeros_like (xpos)

    dx = xedges [1] - xedges [0]
    dy = yedges [1] - yedges [0]
    dz = hist.flatten()

    #cmap = cm.get_cmap('jet') # Get desired colormap - you can change this!
    #cmap = cm.get_cmap('jet')
    cmap = mpl.colormaps['jet']
    max_height = np.max(dz)   # get range of colorbars so we can normalize
    min_height = np.min(dz)
    # scale each z to [0,1], and get their rgb values
    rgba = [cmap((k-min_height)/max_height) for k in dz] 

    ax.bar3d(xpos, ypos, zpos, dx, dy, dz, color=rgba, zsort='average')
    plt.title(title)
    plt.xlabel(xLabel)
    plt.ylabel(yLabel)
    plt.xlim(xymind,xymaxd)
    plt.ylim(xymind,xymaxd)
    #plt.savefig(filename)
    plt.show()
    #plt.clf()

def dsig_dy(xnumber,yy):
        result=1-yy+1/(1-yy)-4*yy/xnumber/(1-yy)+4*pow(yy/xnumber/(1-yy),2)
        return result

def dlum_dy1dy2(xnumber,yy1,yy2):
        result=dsig_dy(xnumber,yy1)*dsig_dy(xnumber,yy2)
        return result

def brute_force(xnumber, nPoints, seed=None):
    ym=xnumber/(xnumber+1)
    dlum_dy1dy2_val_max = dlum_dy1dy2(xnumber,ym,ym)
    nFunctionEval = 0
    yy1_rej_method = []
    yy2_rej_method = []
    maxWeightEncounteredRej=-1.e20
    generator = np.random.RandomState(seed=seed)
    while len(yy1_rej_method) < nPoints:
        rr = generator.uniform(size=3)
        yy1, yy2 = rr[0]*ym, rr[1]*ym
        nFunctionEval += 1
        dlum_dy1dy2_val=dlum_dy1dy2(xnumber,yy1,yy2)
        if(dlum_dy1dy2_val > maxWeightEncounteredRej):
            maxWeightEncounteredRej=dlum_dy1dy2_val
        if(dlum_dy1dy2_val > dlum_dy1dy2_val_max):
            print(' dlum_dy1dy2_val= ',dlum_dy1dy2_val,' exceeds dlum_dy1dy2_val_max= ',dlum_dy1dy2_val_max,' program will now exit ')
            exit(99)
        if(dlum_dy1dy2_val/dlum_dy1dy2_val_max > rr[2]):
            yy1_rej_method.append(yy1)
            yy2_rej_method.append(yy2)
    return {'yy1':yy1_rej_method, 'yy2':yy2_rej_method, 'nFunEval':nFunctionEval, 
            'maxWeightEncountered':maxWeightEncounteredRej, 'dlum_dy1dy2_val_max':dlum_dy1dy2_val_max}

def setup_intervals(xnumber, NN=100, KK=2000, nIterations=4000, alpha_damp=1.5, seed=None):
    """
    Input:
        xnumber: 4 E_{laser gamma} E_{e-}/m_e^2
        NN: Number of intervals in [0, ym], where ym = xnumber/(xnumber+1)
        KK: function evaluations per iteration
        nIterations: number of iterations
        alpha_damp: damping parameter in the Vegas algorithm
    Return:
        Intervals specified by xLow, yLow: each is a 1D numpy array of size NN+1, with
        xLow[0] = 0, xLow[NN] = ym; yLow[0] = 0, yLow[NN] = ym
    """
    ym = xnumber/(xnumber + 1)
    
    # intitial intervals: uniform intervals between 0 and ym
    xLow = ym/NN*np.arange(NN+1)
    delx = np.ones(NN)*ym/NN
    px = np.ones(NN)/ym    # probability density in each interval
    yLow = ym/NN*np.arange(NN+1)
    dely = np.ones(NN)*ym/NN
    py = np.ones(NN)/ym

    generator = np.random.RandomState(seed=seed)
    for _ in range(nIterations):
        ixLow = generator.randint(0, NN, size=KK)
        xx = xLow[ixLow] + delx[ixLow]*generator.uniform(size=KK)
        iyLow = generator.randint(0, NN, size=KK)
        yy = yLow[iyLow] + dely[iyLow]*generator.uniform(size=KK)
        ff = dlum_dy1dy2(xnumber,xx,yy)
        f2barx = np.array([sum((ff[ixLow==i]/py[iyLow[ixLow==i]])**2) for i in range(NN)])
        fbarx = np.sqrt(f2barx)
        f2bary = np.array([sum((ff[iyLow==i]/px[ixLow[iyLow==i]])**2) for i in range(NN)])
        fbary = np.sqrt(f2bary)
        fbardelxSum = np.sum(fbarx*delx)
        fbardelySum = np.sum(fbary*dely)
        logArgx = fbarx*delx/fbardelxSum
        logArgy = fbary*dely/fbardelySum
        mmx = KK*pow((fbarx*delx/fbardelxSum - 1)/np.log(logArgx), alpha_damp)
        mmx = mmx.astype(int)
        mmx = np.where(mmx > 1, mmx, 1)
        mmy = KK*pow((fbary*dely/fbardelySum - 1)/np.log(logArgy), alpha_damp)
        mmy = mmy.astype(int)
        mmy = np.where(mmy > 1, mmy, 1)
        xLowNew = [xLow[i] + np.arange(mmx[i])*delx[i]/mmx[i] for i in range(NN)]
        xLowNew = np.concatenate(xLowNew, axis=0)
        yLowNew = [yLow[i] + np.arange(mmy[i])*dely[i]/mmy[i] for i in range(NN)]
        yLowNew = np.concatenate(yLowNew, axis=0)
        nCombx = int(len(xLowNew)/NN)
        nComby = int(len(yLowNew)/NN)
        i = np.arange(NN)
        xLow[:-1] = xLowNew[i*nCombx]
        yLow[:-1] = yLowNew[i*nComby]
        delx = np.diff(xLow)
        dely = np.diff(yLow)
        px = 1./delx/NN
        py = 1./dely/NN

    return xLow, yLow, delx, dely

def vegas(xnumber, nPoints, vegasRatioFactor, NN=100, KK=2000, nIterations=4000, alpha_damp=1.5, seed=None):
    ym = xnumber/(xnumber +1)
    xLow, yLow, delx, dely = setup_intervals(xnumber, NN, KK, nIterations, alpha_damp, seed)
    vegasRatioMax = vegasRatioFactor*dlum_dy1dy2(xnumber, ym, ym)*NN*NN*delx[NN-2]*dely[NN-2]
    nFunctionEval = 0
    yy1_vegas_method = []
    yy2_vegas_method = []
    yy1_vrho_method = []
    yy2_vrho_method = []
    maxWeightEncountered=-1.e20

    generator = np.random.RandomState(seed=seed)
    while len(yy1_vegas_method) < nPoints:
        ixLow = generator.randint(0, NN)
        xx = xLow[ixLow] + delx[ixLow]*generator.uniform()
        iyLow = generator.randint(0, NN)
        yy = yLow[iyLow] + delx[iyLow]*generator.uniform()
        yy1_vrho_method.append(xx)
        yy2_vrho_method.append(yy)
        nFunctionEval += 1
        dlum_dy1dy2_val = dlum_dy1dy2(xnumber,xx,yy)
        ratio = dlum_dy1dy2_val*NN*NN*delx[ixLow]*dely[iyLow]
        if ratio > maxWeightEncountered:
            maxWeightEncountered = ratio
        if ratio > vegasRatioMax:
            print(' ratio= ',ratio,' exceeds vegasRatioMax= ',vegasRatioMax,' ym= ',ym,' yy= ',yy,' program will now exit ')
            exit(99)
        if ratio/vegasRatioMax > generator.uniform():
            yy1_vegas_method.append(xx)
            yy2_vegas_method.append(yy)

    return {'yy1vrho':yy1_vrho_method, 'yy2vrho':yy2_vrho_method, 
            'yy1vegas':yy1_vegas_method, 'yy2vegas':yy2_vegas_method, 
            'nFunEval':nFunctionEval, 'maxWeightEncountered':maxWeightEncountered, 
            'vegasRatioMax':vegasRatioMax}

def plot_results(xnumber, nPoints, vegasRatioFactor, xymax, nBins=50,  
                 NN=100, KK=2000, nIterations=4000, alpha_damp=1.5, seed=None):
    bf = brute_force(xnumber, nPoints, seed)
    vg = vegas(xnumber, nPoints, vegasRatioFactor, NN, KK, nIterations, alpha_damp, seed)
    ym = xnumber/(xnumber + 1)
    # brute force
    titleRej=r'$\gamma\gamma\ \mathrm{Luminosity\ Rej\ Method}\ x=%.3f,\ ym=%.3f$' %(xnumber, ym)
    lego_plot(bf['yy1'], bf['yy2'],nBins,'y1','y2',0.,xymax,titleRej)
    print(' rejection method xx= ',xnumber,' nPoints= ',nPoints,' nFunctionEval= ',bf['nFunEval'],' maxWeightEncounteredRej= ',bf['maxWeightEncountered'],'  dlum_dy1dy2_val_max= ', bf['dlum_dy1dy2_val_max'])
    plt.show()
    
    # Vegas method
    titleVrho=r'$\gamma\gamma\ \mathrm{Luminosity\ Vrho\ Method}\ x=%.3f,\ ym=%.3f$' %(xnumber, ym)
    lego_plot(vg['yy1vrho'], vg['yy2vrho'], nBins,'y1','y2',0.,xymax,titleVrho)
    print(' vrho method xx= ',xnumber,' nPoints= ',nPoints,' nFunctionEval= ',vg['nFunEval'],' maxWeightEncountered= ',vg['maxWeightEncountered'],' vegasRatioMax= ',vg['vegasRatioMax'],' vegasRatioFactor= ',vegasRatioFactor)
    plt.show()

    titleVegas=r'$\gamma\gamma\ \mathrm{Luminosity\ Vegas\ Method}\ x=%.3f,\ ym=%.3f$' %(xnumber, ym)
    lego_plot(vg['yy1vegas'], vg['yy2vegas'], nBins,'y1','y2',0.,xymax,titleVegas)
    print(' vegas method xx= ',xnumber,' nPoints= ',nPoints,' nFunctionEval= ',vg['nFunEval'])
    plt.show()

plot_results(3.13, 100000, 3.3, 1.01)
#plot_results(1000, 1000, 11, 1.01)